import numpy as np


class dropout:
    """ A dropout layer

    This layer will randomly set elements of a tensor to 0, with some specified probability p.

    Those elements that pass through unmasked are scaled by a factor of  1 / (1 - p), such that
    the magnitude of the tensor, in expectation, is unaffected by the dropout layer.

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.Tensor([1, 2, 3])

    "Dropout" each element of `x` with probability p=0.5. The other elements
    are scaled by 1 / (1-p) so that the magnitude of `x`, on average, remains
    unchanged.

    >>> dropout(x, prob_dropout=0.5)
    Tensor([2., 0., 6.])

    >>> dropout(x, prob_dropout=0.5)
    Tensor([0., 0., 6.])

    >>> dropout(x, prob_dropout=0.5)
    Tensor([2., 4., 0.])
    """

    def __init__(self, prob_dropout):
        """ Parameters
            ----------
            prob_dropout : float
                The probability, specified on [0, 1), that any given element of the input
                will be masked (i.e. set to 0).

                If `prob_dropout=0` the tensor is passed through unchanged."""
        assert 0 <= prob_dropout < 1
        self.p_dropout = prob_dropout

    def __call__(self, x):
        """ Perform the dropout on `x`

        Parameters
        ----------
        x : Union[numpy.ndarray, mygrad.Tensor]
            The data whose elements will each be dropped with probability p.

        Returns
        -------
        Union[numpy.ndarray, mygrad.Tensor]
            `x` with dropout applied.
        """
        if not self.p_dropout:
            return x
        drop_mask = np.random.binomial(1, (1 - self.p_dropout), x.shape)
        return x * drop_mask / (1 - self.p_dropout)

    @property
    def parameters(self):
        """ Access the parameters of the layer.

        Returns
        -------
        Tuple[]
            dropout is a parameterless layer
        """
        return tuple()



from mygrad import maximum, minimum

__all__ = ["leaky_relu"]


def leaky_relu(x, slope):
    """ Returns the leaky rectified linear activation elementwise along x. The leaky ReLU is given
    by max(x, 0) + slope*min(x, 0).

    Parameters
    ----------
    x : mygrad.Tensor
        Input data.

    slope : Union[Real, mygrad.Tensor]
        The slope of the negative activation.

    Returns
    -------
    mygrad.Tensor
        The rectified `x` (elementwise max(x, 0)).
    """
    return maximum(x, 0) + slope * minimum(x, 0)



from mygrad import matmul

from mynn.initializers.uniform import uniform
from mynn.initializers.constant import constant


class dense:
    """ A fully-connected layer.

    This class will perform a dense (fully-connected) linear operation on an (N, D)-shape
    input tensor with a (D, M)-shape weight tensor and a (M,)-shape bias.
    """

    def __init__(
            self,
            input_size,
            output_size,
            *,
            weight_initializer=uniform,
            bias_initializer=constant,
            weight_kwargs=None,
            bias_kwargs=None,
            bias=True,
    ):
        """ Initialize a dense layer.

        Parameters
        ----------
        input_size : int
            The number of features for each input datum.

        output_size : int
            The number of output units (neurons).

        weight_initializer : Callable, optional (default=initializers.uniform)
            The function to use to initialize the weight matrix.

        bias_initializer : Callable, optional (default=initializers.constant)
            The function to use to initialize the bias vector.

        weight_kwargs : Optional[dictionary]
            The keyword arguments to pass to the weight initialization function.

        bias_kwargs : Optional[dictionary]
            The keyword arguments to pass to the bias initialization function.

        bias : bool, optional (default=True)
            If `False` no bias parameter is initialized for the dense layer.
        """
        weight_kwargs = weight_kwargs if weight_kwargs is not None else {}
        bias_kwargs = bias_kwargs if bias_kwargs is not None else {}

        self.weight = weight_initializer(input_size, output_size, **weight_kwargs)
        self.bias = None
        if bias:
            self.bias = bias_initializer(1, output_size, **bias_kwargs)
            self.bias.data = self.bias.data.astype(self.weight.dtype)

    def __call__(self, x):
        """ Perform the forward-pass of the densely-connected layer over `x`.

        Parameters
        ----------
        x : Union[numpy.ndarray, mygrad.Tensor], shape=(N, D)
            The input to pass through the layer.

        Returns
        -------
        mygrad.Tensor
            The result of applying the dense layer wx + b.
        """
        out = matmul(x, self.weight)
        return out + self.bias if self.bias is not None else out

    @property
    def parameters(self):
        """ Access the parameters of the layer.

        Returns
        -------
        Tuple[mygrad.Tensor, mygrad.Tensor]
            The weight and bias of this layer.
        """
        return (self.weight, self.bias) if self.bias is not None else (self.weight,)



from .cross_entropy import softmax_cross_entropy
from .focal_loss import focal_loss
from .huber_loss import huber_loss
from .kl_divergence import kl_divergence
from .l1_loss import l1_loss
from .mean_squared_loss import mean_squared_loss
from .negative_log_likelihood import negative_log_likelihood

__all__ = (
    softmax_cross_entropy,
    focal_loss,
    huber_loss,
    kl_divergence,
    l1_loss,
    mean_squared_loss,
    negative_log_likelihood,
)



import numpy as np


class Adadelta:
    """ Performs the Adadelta optimization procedure from Zeiler.

    Parameters
    ----------
    params : Iterable
        The parameters over which to optimize.

    rho : float ∈ [0, 1), optional (default=0.95)
        The decay rate in the running average.

    eps : float, optional (default=1e-06)
        Epsilon value to better condition the denominator.


    Extended Description
    --------------------
    This optimizer implements the Adadelta optimization procedure described in the paper
      "ADADELTA: An Adaptive Learning Rate Method"
      Matthew Zeiler
    https://arxiv.org/abs/1212.5701
    """

    def __init__(self, params, *, rho=0.95, eps=1e-06):
        self.params = params
        self.rho = rho
        self.eps = eps

        self.g = []
        self.dx = []
        for param in params:
            self.g.append(np.zeros_like(param.data))
            self.dx.append(np.zeros_like(param.data))

    def step(self):
        """ Perform one optimization step.

        This function should be called after accumulating gradients into the parameters of the model
        you wish to optimize via `backward()`. This will perform one step of the adadelta
        optimization algorithm put forward by Zeiler.
        """
        for idx, param in enumerate(self.params):
            grad = param.grad
            if grad is None:
                continue

            self.g[idx] = self.rho * self.g[idx] + (1 - self.rho) * grad ** 2  # step 4
            sqrt_dx = np.sqrt(self.dx[idx] + self.eps)
            sqrt_g = np.sqrt(self.g[idx] + self.eps)
                              
            dx = -sqrt_dx / sqrt_g * grad                                      # step 5
            self.dx[idx] = self.rho * self.dx[idx] + (1 - self.rho) * dx ** 2  # step 6
            param.data += dx                                                   # step 7



import numpy as np


class SGD:
    """ Performs (batched) stochastic gradient descent.

    Parameters
    ----------
    params : Iterable
        The parameters over which to optimize.

    learning_rate : Real, optional (default=0.1)
        The step size.

    momentum : Real ∈ [0, 1), optional (default=0)
        The momentum term.

    weight_decay : Real, optional (default=0)
        The weight decay term.
    """

    def __init__(self, params, *, learning_rate=0.1, momentum=0, weight_decay=0):
        if momentum < 0 or momentum > 1:
            raise ValueError("Momentum must lie within [0, 1)")

        self.params = params
        self.learning_rate = learning_rate
        self.momentum = momentum
        self.weight_decay = weight_decay

        self.param_moments = []
        if momentum != 0:
            for param in params:
                self.param_moments.append(np.zeros_like(param.data))

    def step(self):
        """ Perform one optimization step.

        This function should be called after accumulating gradients into the parameters of the model
        you wish to optimize via `backward()`. This will perform one step of (stochatic) gradient
        descent, optionally with momentum. The update here is given by

          ν = μν - η(dθ)
          θ = θ + ν

        where θ is the parameter list of the model, η is the learning rate, μ is the weighting on
        the momentum term, and ν is the moment of each parameter. Note that if the momentum term is
        0, this simplifies to θ = θ - η(dθ)
        """
        for idx, param in enumerate(self.params):
            if param.grad is None:
                continue

            # no decay on bias
            update = -self.weight_decay * param.data if param.ndim > 1 else 0

            # perform the momentum update
            if self.momentum != 0:
                moment = self.param_moments[idx]
                moment *= self.momentum
                moment -= self.learning_rate * param.grad
                update += moment
            else:
                update += -self.learning_rate * param.grad

            # update parameters
            param.data += update



import numpy as np
from mygrad import Tensor


def he_uniform(*shape, gain=1):
    """ Initialize a :class:`mygrad.Tensor` according to the uniform initialization procedure
    described by He et al.

    Parameters
    ----------
    shape : Sequence[int]
        The shape of the output Tensor. Note that `shape` must be at least two-dimensional.

    gain : Real, optional (default=1)
        The gain (scaling factor) to apply.

    Returns
    -------
    mygrad.Tensor, shape=`shape`
        A Tensor, with values initialized according to the He uniform initialization.

    Extended Description
    --------------------
    He, Zhang, Ren, and Sun put forward this initialization in the paper
        "Delving Deep into Rectifiers: Surpassing Human-Level Performance
        on ImageNet Classification"
    https://arxiv.org/abs/1502.01852

    A Tensor :math:`W` initialized in this way should be drawn from a distribution about

    .. math::
        U[-\sqrt{\frac{6}{(1+a^2)n_l}}, \sqrt{\frac{6}{(1+a^2)n_l}}]

    where :math:`a` is the slope of the rectifier following this layer, which is incorporated
    using the `gain` variable above.
    """
    if len(shape) < 2:
        raise ValueError("He Uniform initialization requires at least two dimensions")

    tensor = np.empty(shape)
    bound = gain / np.sqrt(shape[1] * tensor[0, 0].size) * np.sqrt(3)
    return Tensor(np.random.uniform(-bound, bound, shape))



from .sigmoid import sigmoid


def glu(x, dim=-1):
    """ Returns the Gated Linear Unit A * σ(B), where A and B are split from `x`.

    Parameters
    ----------
    x : mygrad.Tensor
        The input.

    dim : int, optional (default=-1)
        The dimension along which to split the input in half and apply the GLU.

    Returns
    -------
    mygrad.Tensor
        The result of applying the  Gated Linear Unit elementwise to the input.

    Extended Description
    --------------------
    The Gated Linear Unit was proposed in the paper
        "Language Modeling with Gated Convolutional Networks"
        Yann Dauphin, Angela Fan, Michael Auli, David Grangier
    available at https://arxiv.org/abs/1612.08083

    The GLU operation splits the input `x` in half along `dim`, storing the first half in A and the
    second in B. The return value is then A ⊙ σ(B), where ⊙ is elementwise multiplication and σ is
    the sigmoid function.
    """
    first_idx = list(slice(None) for _ in x.shape)
    second_idx = list(slice(None) for _ in x.shape)
    first_idx[dim] = slice(0, x.shape[dim] // 2)
    second_idx[dim] = slice(x.shape[dim] // 2, None)

    first_half = x[tuple(first_idx)]
    second_half = x[tuple(second_idx)]

    if first_half.shape != second_half.shape:
        raise ValueError(
            f"The shapes after splitting must be the same but got {first_half.shape} "
            "and {second_half.shape}"
        )
    return first_half * sigmoid(second_half)



import numpy as np
from mygrad import Tensor

def normal(*shape, mean=0, std=1):
    """ Initialize a :class:`mygrad.Tensor` by drawing from a normal (Gaussian) distribution.

    Parameters
    ----------
    shape : Sequence[int]
        The output shape.

    mean : Real, optional (default=0)
        The mean of the distribution from which to draw.

    std : Real, optional (default=1)
        The standard deviation of the distribution from which to draw.

    Returns
    -------
    mygrad.Tensor, shape=`shape`
        A Tensor, with values drawn from Ɲ(μ, σ²), where μ=`mean` and σ=`std`.
    """
    return Tensor(np.random.normal(mean, std, shape))



from mygrad import abs

__all__ = ["soft_sign"]


def soft_sign(x):
    """ Returns the soft sign function x/(1 + |x|).

    Parameters
    ----------
    x : mygrad.Tensor
        Input data.

    Returns
    -------
    mygrad.Tensor
        The soft sign function applied to `x` elementwise.
    """
    return x / (1 + abs(x))



from mygrad.nnet.activations import softmax as mygrad_softmax

__all__ = ["softmax"]


def softmax(x):
    """ Returns the softmax exp(x) / Σexp(x).

    Parameters
    ----------
    x : mygrad.Tensor
        Input data.

    Returns
    -------
    mygrad.Tensor
        The softmax of `x`.
    """
    return mygrad_softmax(x, constant=False)



from mynn.activations.log_softmax import log_softmax
from mynn.losses.negative_log_likelihood import negative_log_likelihood
from mygrad.nnet.losses import softmax_crossentropy as mygrad_softmax_cross_entropy


def softmax_cross_entropy(x, y, *, weights=None):
    """
    Parameters
    ----------
    x : mygrad.Tensor, shape=(N, C)
        The C class scores for each of the N pieces of data.

    y : Sequence[int]
        The correct class indices, in [0, C), for each datum.

    weights : Sequence[Real], optional (default=None)
        The weighting factor to use on each class, or None.
   
    Returns
    -------
    mygrad.Tensor, shape=()
        The average cross-entropy loss.
    """
    if weights is not None:
        return negative_log_likelihood(log_softmax(x), y, weights=weights)
    return mygrad_softmax_cross_entropy(x, y)



import numpy as np
from mygrad import Tensor


def identity(size):
    """ Initialize the identity matrix of dimension `size`x`size`.

    Parameters
    ----------
    size : int
        The size of the identity matrix.

    Returns
    mygrad.Tensor, shape=(`size`, `size`)
        A Tensor, with ones on the main diagonal and zeros elsewhere.
    """
    return Tensor(np.identity(size))



from mygrad.nnet.activations import logsoftmax as mygrad_log_softmax

__all__ = ["log_softmax"]


def log_softmax(x):
    """ Returns the log softmax of x along its trailing dimension:

        log(exp(x) / Σexp(x))

    However, this implementation is faster and more numerically-stable
    than performing the softmax followed by the log.

    Parameters
    ----------
    x : mygrad.Tensor, shape=(N, C) or shape=(C,)
        Input data.

    Returns
    -------
    mygrad.Tensor, shape=(N,) or shape=()
        The log-softmax of `x`.
    """
    return mygrad_log_softmax(x)



import numpy as np

from mygrad import Tensor
from mygrad.operation_base import Operation

__all__ = ["selu"]


class SELU(Operation):
    """ Returns the scaled exponential linear activation (SELU) elementwise along x. The SELU is
    given by  λɑ(exp(x) - 1) for x < 0 and λx for x ≥ 0.

    Notes
    -----
    The SELU activation was proposed in the paper
        Self-Normalizing Neural Networks
        Günter Klambauer, Thomas Unterthiner, Andreas Mayr, Sepp Hochreiter
    at https://arxiv.org/abs/1706.02515
    """

    def __call__(self, x):
        """
        Parameters
        ----------
        x : mygrad.Tensor
            Input data.

        Returns
        -------
        numpy.ndarray
            The SELU function applied to `x` elementwise.
        """
        self.variables = (x,)

        x = x.data
        self.alpha = 1.6732632423543772848170429916717
        self.scale = 1.0507009873554804934193349852946
        self.exp = self.alpha * (np.exp(x) - 1)
        return self.scale * np.where(x < 0, self.exp, x)

    def backward_var(self, grad, index, **kwargs):
        x = self.variables[index]
        x.backward(
            grad * self.scale * np.where(x.data < 0, self.exp + self.alpha, 1), **kwargs
        )


def selu(x):
    """ Returns the scaled exponential linear activation (SELU) elementwise along x. The SELU is
    given by  λɑ(exp(x) - 1) for x < 0 and λx for x ≥ 0.

    Parameters
    ----------
    x : mygrad.Tensor
        Input data.

    Returns
    -------
    mygrad.Tensor
        The SELU function applied to `x` elementwise.

    Notes
    -----
    The SELU activation was proposed in the paper
        Self-Normalizing Neural Networks
        Günter Klambauer, Thomas Unterthiner, Andreas Mayr, Sepp Hochreiter
    at https://arxiv.org/abs/1706.02515
    """
    return Tensor._op(SELU, x)



from mygrad import Tensor, sum
import numpy as np


def negative_log_likelihood(outputs, targets, *, weights=None):
    """ Returns the (weighted) negative log-likelihood loss between outputs and targets.
    
    Note that this does not compute a softmax, so you should input log-probabilities to this.
    See ``softmax_cross_entropy`` if you need your loss to compute a softmax.

    Parameters
    ----------
    outputs : mygrad.Tensor, shape=(N, C)
        The C log probabilities for each of the N pieces of data.
    
    targets : Union[mygrad.Tensor, Sequence[int]], shape=(N,)
        The correct class indices, in [0, C), for each datum.

    weights : Union[mygrad.Tensor, Sequence[Real]], optional (default=None)
        The weighting factor to use on each class, or None.
    
    Returns
    -------
    mygrad.Tensor, shape=()
        The average (weighted) negative log-likelihood.
    """
    if isinstance(targets, Tensor):
        targets = targets.data

    label_locs = (range(len(targets)), targets)
    factors = weights[targets] if weights is not None else np.ones_like(targets)
    return -sum(outputs[label_locs] * factors) / sum(factors)



from mygrad.operation_base import Operation
from mygrad import Tensor, mean, log
import numpy as np

__all__ = ["softmax_focal_loss", "focal_loss"]


class SoftmaxFocalLoss(Operation):
    """ Returns the focal loss as described in https://arxiv.org/abs/1708.02002 which is
    given by -ɑ(1-p)ˠlog(p).

    Extended Description
    --------------------
    The focal loss is given by

    .. math::
        \frac{1}{N}\sum\limits_{1}^{N}-\alpha \hat{y}_i(1-p_i)^\gamma\log(p_i)

    where :math:`N` is the number of elements in `x` and `y` and :math:`\hat{y}_i` is
    one where :math:`i` is the label of the element :math:`y_i` and 0 elsewhere. That is,
    if the label :math:`y_k` is 1 and there are four possible label values, then
    :math:`\hat{y}_k = (0, 1, 0, 0)`.
    """

    scalar_only = True

    def __call__(self, scores, targets, alpha, gamma):
        """
        Parameters
        ----------
        scores : mygrad.Tensor, shape=(N, C)
            The C class scores for each of the N pieces of data.

        targets : Union[mygrad.Tensor, Sequence[int]], shape=(N,)
            The correct class indices, in [0, C), for each datum.

        alpha : Real
            The ɑ weighting factor in the loss formulation.

        gamma : Real
            The ɣ focusing parameter.

        Returns
        -------
        numpy.ndarray
            The average focal loss.
        """
        if isinstance(targets, Tensor):
            targets = targets.data

        self.variables = (scores,)
        scores = np.copy(scores.data)

        max_scores = np.max(scores, axis=1, keepdims=True)
        np.exp(scores - max_scores, out=scores)
        scores /= np.sum(scores, axis=1, keepdims=True)
        label_locs = (range(len(scores)), targets)

        pc = scores[label_locs]
        one_m_pc = 1 - pc + 1e-14  # correct domain for when gamma < 1 and pc == 1
        log_pc = np.log(pc)

        loss = -np.mean(alpha * one_m_pc ** gamma * log_pc)

        self.back = scores
        self.back[label_locs] -= 1
        deriv = one_m_pc**gamma - pc * gamma * one_m_pc**(gamma - 1) * log_pc
        self.back *= deriv[:, np.newaxis]
        self.back *= alpha / scores.shape[0]
        return loss

    def backward_var(self, grad, index, **kwargs):
        return grad * self.back


def softmax_focal_loss(x, y, *, alpha=1, gamma=0):
    """
    Parameters
    ----------
    outputs : mygrad.Tensor, shape=(N, C)
        The C class scores for each of the N pieces of data.

    targets : Sequence[int], shape=(N,)
        The correct class indices, in [0, C), for each datum.

    alpha : Real, optional (default=1)
        The ɑ weighting factor in the loss formulation.

    gamma : Real, optional (default=0)
        The ɣ focusing parameter. Note that for Ɣ=0 and ɑ=1, this is cross-entropy loss.

    Returns
    -------
    mygrad.Tensor
        The average focal loss.
    """
    return Tensor._op(SoftmaxFocalLoss, x, op_args=(y, alpha, gamma))


def focal_loss(scores, targets, *, alpha=1, gamma=0):
    """ Return the focal loss.

    Parameters
    ----------
    scores : mygrad.Tensor, shape=(N, C)
        The C class scores for each of the N pieces of data.

    targets : Sequence[int], shape=(N,)
        The correct class indices, in [0, C), for each datum.

    alpha : Real, optional (default=1)
        The ɑ weighting factor in the loss formulation.

    gamma : Real, optional (default=0)
        The ɣ focusing parameter. Note that for Ɣ=0 and ɑ=1, this is cross-entropy loss.

    Returns
    -------
    mygrad.Tensor
        The average focal loss.

    Notes
    -----
    This function does not perform a softmax before computing the loss. If you ned to take the 
    softmax before computing the loss, see :class:`SoftmaxFocalLoss` instead.
    """
    if isinstance(targets, Tensor):
        targets = targets.data

    label_locs = (range(len(targets)), targets)
    pc = scores[label_locs]
    return -mean(alpha * (1 - pc + 1e-14) ** gamma * log(pc))



from .constant import constant
from .dirac import dirac
from .glorot_normal import glorot_normal
from .glorot_uniform import glorot_uniform
from .he_normal import he_normal
from .he_uniform import he_uniform
from .identity import identity
from .normal import normal
from .uniform import uniform

__all__ = (
    constant,
    dirac,
    glorot_normal,
    glorot_uniform,
    he_normal,
    he_uniform,
    identity,
    normal,
    uniform,
)



import numpy as np
from mygrad import Tensor


def uniform(*shape, lower_bound=0, upper_bound=1):
    """ Initialize a :class:`mygrad.Tensor` by drawing from a uniform distribution.

    Parameters
    ----------
    shape : Sequence[int]
        The output shape.

    lower_bound : Real, optional (default=0)
        Lower bound on the output interval, inclusive.

    upper_bound : Real, optional (default=1)
        Upper bound on the output interval, exclusive.

    Returns
    -------
    mygrad.Tensor, shape=`shape`
        A Tensor, with values drawn uniformly from [lower_bound, upper_bound).
    """
    return Tensor(np.random.uniform(lower_bound, upper_bound, shape))



from mygrad import maximum, minimum

__all__ = ["hard_tanh"]


def hard_tanh(x, *, lower_bound=-1, upper_bound=1):
    """ Returns the hard hyperbolic tangent function.

    The hard_tanh function is `lower_bound` where `x` <= `lower_bound`, `upper_bound` where 
    `x` >= `upper_bound`, and `x` where `lower_bound` < `x` < `upper_bound`.

    Parameters
    ----------
    x : Union[numpy.ndarray, mygrad.Tensor]
        The input, to which to apply the hard tanh function.

    lower_bound : Real, optional (default=-1)
        The lower bound on the hard tanh.

    upper_bound : Real, optional (default=1)
        The upper bound on the hard tanh.

    Returns
    -------
    mygrad.Tensor
        The result of applpying the hard tanh function elementwise to `x`.
    """
    return maximum(lower_bound, minimum(x, upper_bound))



from mynn.losses.focal_loss import softmax_focal_loss

from mygrad import Tensor, sum, log
from mygrad.nnet.activations import softmax

import numpy as np
from numpy.testing import assert_allclose

from hypothesis import given, assume
import hypothesis.strategies as st
import hypothesis.extra.numpy as hnp


@given(num_datum=st.integers(1, 100),
       num_classes=st.integers(1, 15),
       alpha=st.floats(-1, 1),
       gamma=st.floats(.0, 5),
       data=st.data(),
       grad=st.floats(-1, 1))
def test_focal_loss(num_datum, num_classes, alpha, gamma, data, grad):


    scores = data.draw(hnp.arrays(shape=(num_datum, num_classes),
                                  dtype=float,
                                  elements=st.floats(1, 100)))

    assume((abs(scores.sum(axis=1)) > 0.001).all())

    scores_mygrad = Tensor(scores)
    scores_nn = Tensor(scores)

    truth = np.zeros((num_datum, num_classes))
    targets = data.draw(st.tuples(*(st.integers(0, num_classes - 1) for i in range(num_datum))))

    truth[range(num_datum), targets] = 1

    probs = softmax(scores_mygrad)
    mygrad_focal_loss = sum(truth * (-alpha * (1 - probs + 1e-14)**gamma * log(probs))) / num_datum
    mygrad_focal_loss.backward(grad)

    nn_loss = softmax_focal_loss(scores_nn, targets, alpha=alpha, gamma=gamma)
    nn_loss.backward(grad)

    assert isinstance(nn_loss, Tensor) and nn_loss.ndim == 0
    assert_allclose(nn_loss.data, mygrad_focal_loss.data, atol=1e-4, rtol=1e-4)
    assert_allclose(scores_nn.grad, scores_mygrad.grad, atol=1e-4, rtol=1e-4)

    nn_loss.null_gradients()
    assert scores_nn.grad is None




import numpy as np

from mygrad.nnet.layers.conv import conv_nd

from mynn.initializers.uniform import uniform
from mynn.initializers.constant import constant


class conv:
    """ A convolutional layer.

    This class will perform an n-dimensional convolution on an (N, K, ...)-shape input Tensor
    with a (D, K, ...,)-shape weight Tensor, then add a (D,)-shape bias vector to the result.
    """

    def __init__(
            self,
            input_size,
            output_size,
            *filter_dims,
            stride=1,
            padding=0,
            dilation=1,
            weight_initializer=uniform,
            bias_initializer=constant,
            weight_kwargs=None,
            bias_kwargs=None,
            bias=True,
    ):
        """ Initialize a conv layer.

        Parameters
        ----------
        input_size : int
            The number of feature channels (depth) for each input datum.

        output_size : int
            The number of feature channels (depth) for the output.

        filter_dims : Sequence[int]
            The dimensions of the convolutional filters.

        stride : Union[int, Tuple[int, ...]]
            (keyword-only argument) The step-size with which each
            filter is placed along the H and W axes during the
            convolution. The tuple indicates (stride-0, ...). If a
            single integer is provided, this stride is used for all
            convolved dimensions

        padding : Union[int, Tuple[int, ...]], optional (default=0)
            (keyword-only argument) The number of zeros to be padded
            to both ends of each convolved dimension, respectively.
            If a single integer is provided, this padding is used for
            all of the convolved axes

        dilation : Union[int, Tuple[int, ...]], optional (default=1)
            (keyword-only argument) The spacing used when placing kernel
            elements along the data. E.g. for a 1D convolution the ith
            placement of the kernel multiplied  against the dilated-window:
            `x[:, :, i*s:(i*s + w*d):d]`, where s is
            the stride, w is the kernel-size, and d is the dilation factor.

            If a single integer is provided, that dilation value is used for all
            of the convolved axes.

        weight_initializer : Callable, optional (default=initializers.uniform)
            The function to use to initialize the weight tensor.

        bias_initializer : Callable, optional (default=initializers.constant)
            The function to use to initialize the bias vector.

        weight_kwargs : Optional[dictionary]
            The keyword arguments to pass to the weight initialization function.

        bias_kwargs : Optional[dictionary]
            The keyword arguments to pass to the bias initialization function.

        bias : bool, optional (default=True)
            If `False`, no biar parameter is initialized for the convolutional layer.
        """
        if np.ndim(filter_dims) > 1:  # if the user passes in a Sequence
            filter_dims = filter_dims[0]  # unpack it from the outer Tuple

        weight_kwargs = weight_kwargs if weight_kwargs is not None else {}
        bias_kwargs = bias_kwargs if bias_kwargs is not None else {}

        self.weight = weight_initializer(
            output_size, input_size, *filter_dims, **weight_kwargs
        )
        self.bias = None
        if bias:
            self.bias = bias_initializer(output_size, **bias_kwargs)
            self.bias = self.bias.reshape(1, -1, *(1 for _ in range(len(filter_dims))))
            self.bias.data = self.bias.data.astype(self.weight.dtype)
        self.stride = stride
        self.padding = padding
        self.dilation = dilation

    def __call__(self, x):
        """ Perform the forward-pass of the n-dimensional convolutional layer over `x`.

        Parameters
        ----------
        x : Union[numpy.ndarray, mygrad.Tensor], shape=(N, K, ...)
            The data over which to perform convolution.

        Returns
        -------
        mygrad.Tensor
            The result of convolving `x` with this layer's `weight`, then adding its `bias`.
        """
        out = conv_nd(
            x,
            self.weight,
            stride=self.stride,
            padding=self.padding,
            dilation=self.dilation,
        )
        return out + self.bias if self.bias is not None else out

    @property
    def parameters(self):
        """ Access the parameters of the layer.

        Returns
        -------
        Tuple[mygrad.Tensor, mygrad.Tensor]
            The weight and bias of this layer.
        """
        return (self.weight, self.bias) if self.bias is not None else (self.weight,)



from mygrad.nnet.activations import sigmoid as mygrad_sigmoid

__all__ = ["sigmoid"]


def sigmoid(x):
    """ Returns the sigmoid function of the input elementwise.

    Parameters
    ----------
    x : mygrad.Tensor
        The input.

    Returns
    -------
    mygrad.Tensor
        The input with the sigmoid function applied elementwise.

    Extended Description
    --------------------
    The sigmoid function is given by

    .. math::
        \sigma(x) = \frac{e^x}{e^x + 1}
    """
    return mygrad_sigmoid(x, constant=False)



from .elu import elu
from .glu import glu
from .hard_tanh import hard_tanh
from .leaky_relu import leaky_relu
from .log_softmax import log_softmax
from .relu import relu
from .selu import selu
from .sigmoid import sigmoid
from .softmax import softmax
from .soft_sign import soft_sign


__all__ = (
    elu,
    glu,
    hard_tanh,
    leaky_relu,
    log_softmax,
    relu,
    selu,
    sigmoid,
    softmax,
    soft_sign,
)



import numpy as np
from mygrad import Tensor


def constant(*shape, value=0):
    """ Initialize a :class:`mygrad.Tensor` of shape `shape` with a constant value.

    Parameters
    ----------
    shape : Sequence[int]
        The output shape.

    value : Real, optional (default=0)
        The value with which to fill the Tensor.

    Returns
    -------
    mygrad.Tensor, shape=`shape`
        A Tensor, whose values are `value`.
    """
    return Tensor(np.full(shape, value))



from mygrad import mean


def mean_squared_loss(outputs, targets):
    """ Returns the mean squared error Σ(xᵢ - yᵢ)² over the data points. 

    Parameters
    ----------
    outputs : mygrad.Tensor, shape=(N, any)
        The model outputs, where `N` is the number of items.

    targets : mygrad.Tensor, shape=(N, any)
        The target values, where `N` is the number of items.

    Returns
    -------
    mygrad.Tensor, shape=()
        The mean squared error.

    Extended Description
    --------------------
    The mean squared error is given by

    .. math::
        \frac{1}{N}\sum\limits_{1}^{N}(x_i - y_i)^2

    where :math:`N` is the number of elements in `x` and `y`.
    """
    return mean((outputs - targets) ** 2)



import numpy as np
from mygrad import Tensor


def glorot_uniform(*shape, gain=1):
    """ Initialize a :class:`mygrad.Tensor` according to the uniform initialization procedure
    described by Glorot and Bengio.

    Parameters
    ----------
    shape : Sequence[int]
        The shape of the output Tensor. Note that `shape` must be at least two-dimensional.

    gain : Real, optional (default=1)
        The gain (scaling factor) to apply.

    Returns
    -------
    mygrad.Tensor, shape=`shape`
        A Tensor, with values initialized according to the glorot uniform initialization.

    Extended Description
    --------------------
    Glorot and Bengio put forward this initialization in the paper
        "Understanding the Difficulty of Training Deep Feedforward Neural Networks"
    http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf

    A Tensor :math:`W` initialized in this way should be drawn from a distribution about

    .. math::
        U[-\frac{\sqrt{6}}{\sqrt{n_j+n_{j+1}}}, \frac{\sqrt{6}}{\sqrt{n_j+n_{j+1}}}]
    """
    if len(shape) < 2:
        raise ValueError("Glorot Uniform initialization requires at least two dimensions")

    tensor = np.empty(shape)
    bound = gain * np.sqrt(6 / ((shape[0] + shape[1]) * tensor[0, 0].size))
    return Tensor(np.random.uniform(-bound, bound, shape))



from mygrad import mean, abs


def l1_loss(outputs, targets):
    """ Returns the L¹ loss Σ|xᵢ - yᵢ| averaged over the number of data points. 

    Parameters
    ----------
    outputs : mygrad.Tensor, shape=(N,)
        The predictions for each of the N pieces of data.

    targets : numpy.ndarray, shape=(N,)
        The correct value for each of the N pieces of data.

    Returns
    -------
    mygrad.Tensor, shape=()
        The average L¹ loss.

    Extended Description
    --------------------
    The L1 loss is given by
    
    .. math::
        \frac{1}{N}\sum\limits_{1}^{N}|x_i - y_i|

    where :math:`N` is the number of elements in `x` and `y`.
    """
    return mean(abs(outputs - targets))



from mygrad import Tensor

from mynn.activations.leaky_relu import leaky_relu


class prelu:
    """ A parametric rectified linear unit.

    This class maintains the learned slope parameter of a PReLU unit, which is a leaky ReLU with
    learned slope in the negative region.

    """

    def __init__(self, slope=0.1):
        """ Parameters
            ----------
            slope : Real, optional (default=0.1)
                The initial value to use for the slope."""
        self.slope = Tensor(slope)

    def __call__(self, x):
        """ Forward the input through the PReLU.

        Parameters
        ----------
        x : Union[numpy.ndarray, mygrad.Tensor]
            The data, to which to apply the PReLU.

        Returns
        -------
        mygrad.Tensor
            The result of applying the PReLU elementwise across `x`.
        """
        return leaky_relu(x, self.slope)

    @property
    def parameters(self):
        """ Access the parameters of the layer.

        Returns
        -------
        Tuple[mygrad.Tensor]
            The slope of the PReLU unit.
        """
        return (self.slope,)



from mygrad import mean, log


def kl_divergence(outputs, targets):
    """ Returns the Kullback-Leibler divergence loss from the outputs to the targets.
    
    The KL-Divergence loss for a single sample is given by yᵢ⊙(log(yᵢ) - xᵢ)

    Parameters
    ----------
    outputs : mygrad.Tensor, shape=(N, any)
        The model outputs for each of the N pieces of data.

    targets : numpy.ndarray, shape=(N, any)
        The correct vaue for each datum.

    Returns
    -------
    mygrad.Tensor, shape=()
        The mean Kullback-Leibler divergence.
    """
    return mean(targets * (log(targets) - outputs))



import numpy as np

from tests.wrappers.uber import backprop_test_factory, fwdprop_test_factory
from mynn.activations import (
    elu,
    glu,
    hard_tanh,
    leaky_relu,
    log_softmax,
    relu,
    selu,
    sigmoid,
    softmax,
    soft_sign,
)


def _away_from_zero(*args, **kwargs):
    x = args[0]
    return np.all(np.abs(x.data) > 1e-08)


def test_elu_fwd():
    pass


def test_elu_bkwd():
    pass


def test_glu_fwd():
    pass


def test_glu_bkwd():
    pass


def test_hard_tanh_fwd():
    pass


def test_hard_tanh_bkwd():
    pass


def _np_leaky_relu(x, slope):
    return np.maximum(x, 0) + slope * np.minimum(x, 0)


@fwdprop_test_factory(mygrad_func=leaky_relu, true_func=_np_leaky_relu, num_arrays=1)
def leaky_relu_fwd():
    pass


@backprop_test_factory(mygrad_func=leaky_relu, true_func=_np_leaky_relu, num_arrays=1)
def leaky_relu_bkwd():
    pass


# @fwdprop_test_factory(
#     mygrad_func=log_softmax,
#     true_func=lambda x: np.log(np.exp(x) / np.exp(x).sum()),
#     num_arrays=1,
#     assumptions=_away_from_zero,
# )
# def test_log_softmax_fwd():
#     pass


def _np_log_softmax(x):
    x = x - x.max()
    x = np.exp(x)
    return np.log(x / x.sum())

# @backprop_test_factory(
#     mygrad_func=log_softmax,
#     true_func=_np_log_softmax,
#     num_arrays=1,
#     assumptions=_away_from_zero,
# )
# def test_log_softmax_bkwd():
#     pass


@fwdprop_test_factory(
    mygrad_func=relu, true_func=lambda x: np.maximum(x, 0), num_arrays=1
)
def test_relu_fwd():
    pass


@backprop_test_factory(
    mygrad_func=relu,
    true_func=lambda x: np.maximum(x, 0),
    num_arrays=1,
    assumptions=_away_from_zero,
)
def test_relu_bkwd():
    pass


def test_selu_fwd():
    pass


def test_selu_bkwd():
    pass


def test_sigmoid_fwd():
    pass


def test_sigmoid_bkwd():
    pass


def test_softmax_fwd():
    pass


def test_softmax_bkwd():
    pass


def test_soft_sign_fwd():
    pass


def test_soft_sign_bkwd():
    pass



import numpy as np


class Adam:
    """ Performs the Adaptive Moment Estimation optimization procedure from Kingma and Ba.

    Parameters
    ----------
    params : Iterable
        The parameters over which to optimize.

    learning_rate : float, optional (default=0.001)
        Alpha in the original algorithm, this is the step size.

    beta1 : float ∈ [0, 1), optional (default=0.9)
        Decay rate for the first moment estimate (mean of the gradient).

    beta2 : float ∈ [0, 1), optional (default=0.999)
        Decay rate for the second moment estimate (uncentered variance of the gradient).

    eps : float, optional (default=1e-08)
        Epsilon value to prevent divide-by-zero.
        
    weight_decay : Real, optional (default=0)
        The weight decay term.

    Extended Description
    --------------------
    This optimizer implements the Adam optimization procedure described in the paper
      "Adam: A Method for Stochastic Optimization"
      Diederik P. Kingma and Jimmy Ba
    https://arxiv.org/abs/1412.6980
    """

    def __init__(
            self,
            params,
            *,
            learning_rate=0.001,
            beta1=0.9,
            beta2=0.999,
            eps=1e-08,
            weight_decay=0,
    ):
        assert 0 <= beta1 < 1
        assert 0 <= beta2 < 1
        self.params = params
        self.learning_rate = learning_rate
        self.beta1 = beta1
        self.beta2 = beta2
        self.eps = eps
        self.t = 0  # timestep
        self.m = []  # first moment estimate
        self.v = []  # second moment estimate
        for param in params:
            self.m.append(np.zeros_like(param.data))
            self.v.append(np.zeros_like(param.data))
        self.weight_decay = weight_decay

    def step(self):
        """ Perform one optimization step.

        This function should be called after accumulating gradients into the parameters of the model
        you wish to optimize via `backward()`. This will perform one step of the Adam optimization
        algorithm put forward by Kingma and Ba.
        """
        self.t += 1
        for idx, param in enumerate(self.params):
            if param.grad is None:
                continue

            self.m[idx] = self.beta1 * self.m[idx] + (1 - self.beta1) * param.grad
            self.v[idx] = self.beta2 * self.v[idx] + (1 - self.beta2) * param.grad ** 2

            m_hat = self.m[idx] / (1 - self.beta1 ** self.t)
            v_hat = self.v[idx] / (1 - self.beta2 ** self.t)

            update = -self.weight_decay * param.data if param.ndim > 1 else 0
            update += -self.learning_rate * m_hat / (np.sqrt(v_hat) + self.eps)
            param.data += update



import numpy as np


class AdaGrad:
    """ Performs the AdaGrad optimization procedure from Duchi, Hazan, and Singer.

    Parameters
    ----------
    params : Iterable
        The parameters over which to optimize.

    learning_rate : float, optional (default=0.01)
        The step size; eta in the paper.

    Extended Description
    --------------------
    This optimizer implements the AdaGrad procedure described in the paper
      "Adaptive Subgradient Methods for Online Learning and Stochastic Optimization"
      John Duchi, Elad Hazan, and Yoram Singer
    http://jmlr.org/papers/volume12/duchi11a/duchi11a.pdf
    """

    def __init__(self, params, *, learning_rate=0.01):
        raise NotImplementedError



from mygrad.nnet.activations import relu as mygrad_relu

__all__ = ["relu"]


def relu(x):
    """ Returns the rectified linear activation max(x, 0) elementwise along x.

    Parameters
    ----------
    x : mygrad.Tensor
        Input data.

    Returns
    -------
    mygrad.Tensor
        The rectified `x` (elementwise max(x, 0)).
    """
    return mygrad_relu(x, constant=False)



from setuptools import setup, find_packages

import versioneer


DISTNAME = 'mynn'
DESCRIPTION = 'A pure-Python neural network library'
LICENSE = 'MIT'
AUTHOR = 'David Mascharka'
AUTHOR_EMAIL = 'davidmascharka@gmail.com'
URL = 'https://github.com/davidmascharka/MyNN'
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Education",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Topic :: Scientific/Engineering",
]
INSTALL_REQUIRES = ['numpy >= 1.13', 'mygrad >= 1.0']
TESTS_REQUIRE = ['pytest >= 3.8', 'hypothesis >= 4.6']
LONG_DESCRIPTION = """
MyNN is a simple NumPy-centric neural network library that builds on top of MyGrad. It provides
convenient wrappers for such functionality as

- Convenient neural network layers (e.g. convolutional, dense, batch normalization, dropout)
- Weight initialization functions (e.g. Glorot, He, uniform, normal)
- Neural network activation functions (e.g. elu, glu, tanh, sigmoid)
- Common loss functions (e.g. cross-entropy, KL-divergence, Huber loss)
- Optimization algorithms (e.g. sgd, adadelta, adam, rmsprop)

MyNN comes complete with several examples to ramp you up to being a fluent user of the library.
It was written as an extension to MyGrad for rapid prototyping of neural networks with minimal dependencies,
a clean codebase with excellent documentation, and as a learning tool.
"""

if __name__ == '__main__':
    setup(
        name=DISTNAME,
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        license=LICENSE,
        author=AUTHOR,
        classifiers=CLASSIFIERS,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_ocntent_type="text/markdown",
        url=URL,
        python_requires=">=3.6",
        packages=find_packages(),
        install_requires=INSTALL_REQUIRES,
        tests_require=TESTS_REQUIRE,
    )



import numpy as np
import mygrad as mg

from mygrad.nnet.layers.batchnorm import batchnorm as tensor_batchnorm


class batchnorm:
    """ A batch normalization layer.

    This class will perform an n-dimensional batch normalization operation on an
    (N, D, ...)-shaped tensor scaled by γ of shape (D, ...) and shifted by β of shape (D, ...).
    """

    def __init__(self, input_channels, momentum=0.1):
        """ Initialize a batch normalization layer.

        Parameters
        ----------
        input_channels : int
            The number of channels of the data to be batch-normalized.

        momentum : float, optional (default=0.1)
            The momentum value used to maintain moving averages.
        """
        self.gamma = mg.ones((1, input_channels), dtype=np.float32)
        self.beta = mg.zeros((1, input_channels), dtype=np.float32)

        self.moving_mean = np.zeros((1, input_channels), dtype=np.float32)
        self.moving_variance = np.zeros((1, input_channels), dtype=np.float32)

        self.momentum = momentum

        self.input_channels = input_channels

    def __call__(self, x, test=False):
        """ Perform the forward-pass of n-dimensional batch normalization over axis 1 on `x`.

        Parameters
        ----------
        x : Union[numpy.ndarray, mygrad.Tensor], shape=(N, D, ...)
            The data to normalize.

        test : boolean, optional (default=False)
            Determines whether the layer is being used at training time. The mean and variance
            will be computed for the batch during training, while averaged batch statistics will
            be used at test time.
        """
        if test:
            keepdims_shape = tuple(1 if n != 1 else d for n, d in enumerate(x.shape))
            x = x - self.moving_mean.reshape(keepdims_shape)
            x /= np.sqrt(self.moving_variance.reshape(keepdims_shape) + 1e-08)
            return self.gamma * x + self.beta

        x_norm = tensor_batchnorm(x, gamma=self.gamma, beta=self.beta, eps=1e-08)

        batch_mean = x_norm.creator.mean
        batch_variance = x_norm.creator.var

        self.moving_mean *= 1 - self.momentum
        self.moving_mean += self.momentum * batch_mean
        self.moving_variance *= 1 - self.momentum
        self.moving_variance += self.momentum * batch_variance
        return x_norm

    @property
    def parameters(self):
        """ Access the learnable parameters of the layer.

        Returns
        -------
        Tuple[mygrad.Tensor, mygrad.Tensor]
            The gamma and beta values of this layer.
        """
        return (self.gamma, self.beta)



from .adadelta import Adadelta
from .adagrad import AdaGrad
from .adamax import AdaMax
from .adam import Adam
from .rmsprop import RMSprop
from .sgd import SGD

__all__ = (Adadelta, AdaGrad, AdaMax, Adam, RMSprop, SGD)



import numpy as np
from mygrad import Tensor
from mygrad.operation_base import Operation

__all__ = ["elu"]


class ELU(Operation):
    """ Returns the exponential linear activation (ELU) elementwise along x. The ELU is given by 
    ɑ(exp(x) - 1) for x < 0 and x for x ≥ 0.
    """

    def __call__(self, x, alpha):
        """
        Parameters
        ----------
        x : mygrad.Tensor
            Input data.

        alpha : Real
            The multiplicative factor on the negative activation.

        Returns
        -------
        numpy.ndarray
            The ELU function applied to `x` elementwise.
        """
        self.variables = (x,)

        x = x.data
        self.exp = alpha * (np.exp(x) - 1)
        self.alpha = alpha
        return np.where(x < 0, self.exp, x)

    def backward_var(self, grad, index, **kwargs):
        x = self.variables[index]
        x.backward(grad * np.where(x.data < 0, self.exp + self.alpha, 1), **kwargs)


def elu(x, alpha):
    """ Returns the exponential linear activation (ELU) elementwise along x. The ELU is given by 
    ɑ(exp(x) - 1) for x < 0 and x for x ≥ 0.

    Parameters
    ----------
    x : mygrad.Tensor
        Input data.

    alpha : Real
        The multiplicative factor on the negative activation.

    Returns
    -------
    mygrad.Tensor
        The ELU function applied to `x` elementwise.
    """
    return Tensor._op(ELU, x, op_args=(alpha,))



import numpy as np


class RMSprop:
    """ Performs the rmsprop optimization procedure from Hinton.

    Parameters
    ----------
    params : Iterable
        The parameters over which to optimize.

    learning_rate : float, optional (default=0.01)
        The step size.

    alpha : float ∈ [0, 1), optional (default=0.9)
        A smoothing factor.

    eps : float, optional (default=1e-08)
        Epsilon term to improve numerical stability.

    Extended Description
    --------------------
    This optimizer implements the rmsprop procedure described in the lecture notes
      "Lecture 6e rmsprop: Divide the gradient by a running average of its recent magnitude
      Geoff Hinton with Nitish Srivastava and Kevin Swersky
    http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf
    """

    def __init__(self, params, *, learning_rate=0.01, alpha=0.9, eps=1e-08):
        raise NotImplementedError



import numpy as np

from mygrad.operation_base import Operation
from mygrad import Tensor

__all__ = ["huber_loss"]


class HuberLoss(Operation):
    """ Returns the Huber loss (smooth L1).

    Extended Description
    --------------------
    The Huber loss is given by

    .. math::
        L_\delta(x, y) = \frac{1}{N}\sum\limits_1^N \bigl\{ \begin{array}{l l} 
            \frac{(x_i - y_i)^2}{2} & |x_i - y_i| \leq \delta\\
            \delta|x_i - y_i| - \frac{\delta}{2} & |x_i - y_i| > \delta\end{array}
    """

    scalar_only = True

    def __call__(self, outputs, targets, delta=1):
        """
        Parameters
        ----------
        outputs : mygrad.Tensor, shape=(N, any)
            The output for each of the N pieces of data.

        targets : Union[mygrad.Tensor, numpy.ndarray], shape=(N, any)
            The target for each datum.

        delta : Real > 0, optional (default=1)
            The value under which to use a squared error.

        Returns
        -------
        numpy.ndarray
            The average Huber loss.
        """
        if delta <= 0:
            raise ValueError("Delta must be > 0")

        if isinstance(targets, Tensor):
            targets = targets.data

        self.variables = (outputs,)
        outs = outputs.data

        diff = outs - targets
        sign = np.sign(diff)
        np.abs(diff, out=diff)

        loss = diff.copy()
        np.multiply(loss, diff / 2, where=(diff < 1), out=loss)
        loss[diff >= 1] -= 0.5 * delta
        loss[diff >= 1] *= delta

        self.back = np.where(diff < 1, outs - targets, delta * sign) / outs.size

        return np.mean(loss)

    def backward_var(self, grad, index, **kwargs):
        return grad * self.back


def huber_loss(x, y, *, delta=1):
    """ Returns the Huber loss (smooth L1).

    Parameters
    ----------
    outputs : mygrad.Tensor, shape=(N, any)
        The output for each of the N pieces of data.

    targets : Union[mygrad.Tensor, numpy.ndarray], shape=(N, any)
        The target for each datum.

    delta : Real > 0, optional (default=1)
        The value under which to use a squared error.

    Returns
    -------
    mygrad.Tensor, shape=()
        The average Huber loss.

    Extended Description
    --------------------
    The Huber loss is given by

    .. math::
        L_\delta(x, y) = \frac{1}{N}\sum\limits_1^N \bigl\{ \begin{array}{l l} 
            \frac{(x_i - y_i)^2}{2} & |x_i - y_i| \leq \delta\\
            \delta|x_i - y_i| - \frac{\delta}{2} & |x_i - y_i| > \delta\end{array}

    """
    return Tensor._op(HuberLoss, x, op_args=(y, delta))
